/* suid root'able version of 'chroot'.  See 'info chroot'.
 * 
 * This binary has to be suid root as only root can use chroot() call.
 * 1. After that has been done, the real user ID and GID have to be
 *    restored so that user can't run anything with root priviledges.
 * 2. Then the program user gave, will be run as chrooted.
 * 
 * Copyright (c) 2002 Eero Tamminen <eero.tamminen@creanor.com>
 *
 * chroot-uid is licensed under GPL.
 * You can read the full license at http://www.gnu.org/licenses/gpl.txt
 */
#include <stdio.h>
#include <unistd.h>
#include <sys/types.h>
#include <errno.h>
#include <stdlib.h>
#include <grp.h>

int main(int argc, char *argv[], char *envp[])
{
  char *path, *prog, **args;
  int uid, gid;

  /* check argument validity */
  if (argc < 3 || argv[1][0] != '/') {
    fprintf(stderr, "usage: sudo %s   /absolute-path  path/program\n", *argv);
    fprintf(stderr, "  See 'man chroot'.  This is a version that can be run sudo root\n");
    fprintf(stderr, "  as it restores real user ID before running the program.\n");
    fprintf(stderr, "  Path to the program should be relative to first path!\n");
    return -1;
  }
  path = argv[1];
  prog = argv[2];
  args = &(argv[2]);
  
  /* change to given directory and chroot to it */
  if (chdir(path) < 0) {
    perror("ERROR");
    fprintf(stderr, "chdir(%s)\n", path);
    return -1;
  }
  if (chroot(path) < 0) {
    perror("ERROR");
    fprintf(stderr, "chroot(%s)\n", path);
    return -1;
  }
  
  /* read _real_ user ID  and group and restore them */

  uid = strtol(getenv("SUDO_UID"),NULL,0);
  if (errno != 0) {
	  perror("ERROR");
	  return -1;
  }
  gid = strtol(getenv("SUDO_GID"),NULL,0);
  if (errno != 0) {
	  perror("ERROR");
	  return -1;
  }

  if (initgroups(getenv("SUDO_USER"), gid) < 0) {
	  perror("ERROR");
	  return -1;
  }
  if (setgid(gid) < 0) {
    perror("ERROR");
    fprintf(stderr, "ERROR: setgid()\n");
    return -1;
  }
  if (setuid(uid) < 0) {
    perror("ERROR");
    fprintf(stderr, "ERROR: setuid()\n");
    return -1;
  }

  /* run the given program */
  if (execve(prog, args, envp) < 0) {
    perror("ERROR");
    fprintf(stderr, "ERROR: execve(%s, %s..., %s...)\n", prog, *args, *envp);
    return -1;
  }
  return 0;
}
